/*
 * display.c
 *
 * Created: 2017-07-16 21:49:22
 *  Author: doom
 */ 

#define F_CPU 11059200
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <inttypes.h>
#include <avr/pgmspace.h>
#include "display.h"
#include "defs.h"

//---------------------------------------------------
// Komendy sterujce wywietlaczem
#define LCDC_CLS				0x01
#define LCDC_HOME				0x02
#define LCDC_MODE				0x04
	#define LCDC_MODER			0x02
	#define LCDC_MODEL			0
	#define LCDC_MODEMOVE		0x01
#define LCDC_ON 				0x08
	#define LCDC_ONDISPLAY		0x04
	#define LCDC_ONCURSOR		0x02
	#define LCDC_ONBLINK		0x01
#define LCDC_SHIFT				0x10
	#define LCDC_SHIFTDISP		0x08
	#define LCDC_SHIFTR			0x04
	#define LCDC_SHIFTL			0
#define LCDC_FUNC				0x20
	#define LCDC_FUNC8b			0x10
	#define LCDC_FUNC4b			0
	#define LCDC_FUNC2L			0x08
	#define LCDC_FUNC1L			0
	#define LCDC_FUNC5x10		0x4
	#define LCDC_FUNC5x7		0
#define LCDC_CGA				0x40
#define LCDC_DDA				0x80

//---------------------------------------------------
#define MAX_SPEC_CHARS			8
const char empty[8] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
const char degree[8] = {0x06, 0x09, 0x09, 0x06, 0x00, 0x00, 0x00, 0x00};
const char cdegree[8] = {0x00, 0x18, 0x1B, 0x04, 0x04, 0x04, 0x03, 0x00};
const char maxSymbol[8] = {0x1F, 0x04, 0x0E, 0x15, 0x04, 0x04, 0x04, 0x00};	
const char minSymbol[8] = {0x04, 0x04, 0x04, 0x15, 0x0E, 0x04, 0x1F, 0x00};		

//------------------------------------------------------------------
struct {
	uint8_t rows;
	uint8_t cols;
	uint8_t mode;
	uint8_t bkl;
} dispConfig;

//------------------------------------------------------------------
struct {
	uint8_t buff[DISP_BUFF_LENGTH];
	uint16_t index;
	uint8_t refresh;
} display;

//------------------------------------------------------------------
struct {
	uint8_t buff[DISP_BUFF_LENGTH];
	uint16_t countdown;
} dispMessage;

//------------------------------------------------------------------
static void LCDCommand(char command);
static void LCDData(char data);
static void LCDClear(void);
static void LCDHome(void);
static void LCDLocate(char lcd_y, char lcd_x);
static void LCDSendHalf(char byte);
static void LCDSendByte(char data);
static void LCDSetSpecChar(char spec, char* pData);

static void PWMInit(void);

//------------------------------------------------------------------
void DispInit(void){	
	PWMInit();
	DispConfigSet(LCD_NUM_OF_ROWS, LCD_NUM_OF_COLS, 0);	
	DispBuffClear();
	DispBuffSetPos(0,0);
	LCDInit();
}

//---------------------------------------------------
static void PWMInit(void){
	LCD_CTRL_PORT |= (1<<LCD_CTRL_BKL);
	LCD_CTRL_DDR |= (1<<LCD_CTRL_BKL);

/*	ICR1 = 0xFF;
	OCR1A = 0;
	OCR1B = 0xFF;
	TCCR1A = (1<<COM1A1 | 1<<COM1A0 | 1<<COM1B1 | 1<<COM1B0 | 1<<WGM11 | 0<<WGM10);	
	TCCR1B = (1<< WGM13 | 1<<WGM12 | 1<<CS10);	
*/
}

//---------------------------------------------------
void DispBacklightSet(uint8_t backlight){
	if (backlight > 0){
		LCD_CTRL_PORT &= ~(1<<LCD_CTRL_BKL);
	} else {
		LCD_CTRL_PORT |= (1<<LCD_CTRL_BKL);		
	}
	/*dispConfig.bkl = backlight;
	OCR1A = dispConfig.bkl;
	TCCR1A |= (1<<COM1A1 | 1<<COM1A0);				//FAST PWM
	*/
}

//------------------------------------------------------------------
void DispConfigSet(uint8_t rows, uint8_t cols, uint8_t mode){
	dispConfig.rows = rows;
	dispConfig.cols = cols;
	dispConfig.mode = mode;
}

//------------------------------------------------------------------
uint8_t DispBuffAdd(uint8_t *data, uint8_t stopChar, uint8_t len) {
	uint8_t status = 0;
	uint8_t i = 0;

	while (status == 0) {
		if (display.index < DISP_BUFF_LENGTH) {
			if (len > 0) {
				if (i >= len) status = 1;
			} else {
				if (*data == stopChar) status = 1;
			}
			if (status == 0) {
				display.buff[display.index] = *data;
				display.index++;
				data++;
				i++;
				display.refresh = 1;
			}
		} else status = 2;
	}
	return status;
}

//------------------------------------------------------------------
uint8_t DispBuffAddPgm(uint8_t *data, uint8_t stopChar, uint8_t len) {
	uint8_t status = 0;
	uint8_t i = 0;

	while (status == 0) {
		if (display.index < DISP_BUFF_LENGTH) {
			if (data == 0) status = 3;
			if (len > 0) {
				if (i >= len) status = 1;
			} else {
				if (pgm_read_byte(data) == stopChar) status = 1;
			}
			if (status == 0) {
				display.buff[display.index] = pgm_read_byte(data);
				display.index++;
				data++;
				i++;
				display.refresh = 1;
			}
		} else status = 2;
	}
	return status;
}

//------------------------------------------------------------------
uint8_t DispBuffAddByte(uint8_t data) {
	return DispBuffAdd(&data,0,1);
}

//---------------------------------------------------
uint8_t DispBuffAddTextLen(uint8_t *data, uint16_t len) {
	return DispBuffAdd(data,0,len);
}

//------------------------------------------------------------------
void DispBuffSetPos(uint8_t y, uint8_t x) {
	if ((x < dispConfig.cols) && (y < dispConfig.rows)) {
		display.index = ((y *dispConfig.cols) + x);
	}
}

//------------------------------------------------------------------
void DispBuffSetIndexPos(uint8_t i) {
	display.index = i;
}

//---------------------------------------------------
uint8_t DispBuffAddText(uint8_t *data) {
	return DispBuffAdd(data,0,0);
}

//---------------------------------------------------
uint8_t DispBuffAddPgmText(uint8_t *data) {
	return DispBuffAddPgm(data,0,0);
}

//---------------------------------------------------
uint8_t DispBuffAddPosText(uint8_t *data, uint8_t y, uint8_t x) {
	DispBuffSetPos(y, x);
	return DispBuffAdd(data,0,0);
}

//---------------------------------------------------------------------------------
void DispBuffAddValue(int16_t val, uint8_t dotpos, uint8_t const_space){
	uint8_t visible = 0;
	uint8_t d;
	uint8_t status = 0;
	uint16_t div = 10000;


	if (dotpos > 0) div /= 10;

	if (val < 0) {
		DispBuffAddByte('-');
		val *= -1;
	} else {
		if (const_space > 0) DispBuffAddByte(' ');
	}

	//if(val > 9999) val = 9999;
	while (status == 0)
	{
		d = (val / div);
		if (d) visible = 1;
		if (div == 1) visible = 1;
		if (dotpos > 0) {
			if ((div / 10) == dotpos) visible = 1;
		}

		if (visible) {
			DispBuffAddByte(d + '0');
		} else {
			if (const_space > 0) DispBuffAddByte(' ');
		}

		if (div == 1) status = 1;
		if (dotpos > 0) {
			if ((div / 10) == dotpos) DispBuffAddByte('.');
		}

		val = (val % div);
		div = (div / 10);
	}
}

//---------------------------------------------------
void DispBuffClear(void) {
	uint8_t i;
	for (i=0; i<DISP_BUFF_LENGTH; i++){
		display.buff[i] = ' ';
	}
}

//------------------------------------------------------------------
void DispNewMessage(uint8_t *src, uint8_t time) {
	uint8_t i = 0;

	for (i=0; i < DISP_BUFF_LENGTH; i++){
		if ((*src > 0) && (src > 0)){
			dispMessage.buff[i] = *src;
			src++;
		} else {
			dispMessage.buff[i] = ' ';
		}
	}
	dispMessage.countdown = time;
	display.refresh = 1;
}


//---------------------------------------------------
void DispBuffSend(uint8_t* src){
	uint8_t i = 0;
	uint8_t row = 0;
	uint8_t status = 0;

	while(status == 0){
		row = (i / dispConfig.cols);
		if (row < dispConfig.rows){
			
			if ((i % dispConfig.cols) == 0) {
				if (dispConfig.mode == 0){
					if (row < 2) {
						LCDLocate(row,0);
					} else {
						LCDLocate(row,dispConfig.cols);						
					}
				}
				if (dispConfig.mode == 1){
					LCDLocate(row,0);			
				}
			}

			//LCDData(display.buff[i]);	i++;
			LCDData(*src);	src++;
			i++;
			if (i >= DISP_BUFF_LENGTH) status = 1;
		} else {
			status = 1;
		}
	}
}

//---------------------------------------------------
void DispRefresh(void) {
	display.refresh = 1;
}

//---------------------------------------------------
void DispExe(void) {
	if (dispMessage.countdown > 0){
		dispMessage.countdown--;
		if (dispMessage.countdown == 0){
			display.refresh = 1;
		}
	}
}

//---------------------------------------------------
void DispProcess(void) {

	if (display.refresh > 0) {
		display.refresh = 0;
		if (dispMessage.countdown == 0) {
			DispBuffSend(&display.buff[0]);
		} else {
			DispBuffSend(&dispMessage.buff[0]);			
		}
	}
}

//---------------------------------------------------
uint8_t DispMessageCheck(void){
	if (dispMessage.countdown > 0) return 1;
	return 0;
}

//---------------------------------------------------
void DispReset(void) {
	display.refresh = 1;
	LCDInit();
	LCDSetSpecChar(0, empty);
	LCDSetSpecChar(1, degree);	
}

//---------------------------------------------------
void LCDInit(void){
	//inits ports
	LCD_DATA_PORT |= (1<<LCD_DB4 | 1<<LCD_DB5 | 1<<LCD_DB6 | 1<<LCD_DB7);
	LCD_COM_PORT  |= (1<<LCD_EN | 1<<LCD_RS);
	LCD_DATA_DDR  |= (1<<LCD_DB4 | 1<<LCD_DB5 | 1<<LCD_DB6 | 1<<LCD_DB7);
	LCD_COM_DDR   |= (1<<LCD_EN | 1<<LCD_RS);
	_delay_ms(10);
	LCD_COM_PORT &= ~(1<<LCD_RS);
	_delay_ms(20);
	LCDSendHalf((LCDC_FUNC|LCDC_FUNC8b) >> 0x04);
	_delay_ms(10);
	LCDSendHalf((LCDC_FUNC|LCDC_FUNC8b) >> 0x04);
	_delay_ms(2);
	LCDSendHalf((LCDC_FUNC|LCDC_FUNC8b) >> 0x04);
	_delay_ms(2);
	LCDSendHalf((LCDC_FUNC|LCDC_FUNC4b) >> 0x04);
	_delay_ms(2);

	LCDCommand(LCDC_FUNC|LCDC_FUNC4b|LCDC_FUNC2L|LCDC_FUNC5x7);
	LCDCommand(LCDC_ON);
	LCDClear();
	LCDCommand(LCDC_MODE|LCDC_MODER);
	LCDCommand(LCDC_ON|LCDC_ONDISPLAY);
	
	LCDSetSpecChar(0, empty);	
	LCDSetSpecChar(1, degree);
	LCDSetSpecChar(2, cdegree);
	LCDSetSpecChar(3, maxSymbol);
	LCDSetSpecChar(4, minSymbol);
}

//---------------------------------------------------
static void LCDSendHalf(char byte){
	LCD_COM_PORT |= (1<<LCD_EN);
	
	if (byte & (1<<0))
	LCD_DATA_PORT |= (1<<LCD_DB4);
	else
	LCD_DATA_PORT &= ~(1<<LCD_DB4);
	
	if (byte & (1<<1))
	LCD_DATA_PORT |= (1<<LCD_DB5);
	else
	LCD_DATA_PORT &= ~(1<<LCD_DB5);

	if (byte & (1<<2))
	LCD_DATA_PORT |= (1<<LCD_DB6);
	else
	LCD_DATA_PORT &= ~(1<<LCD_DB6);

	if (byte & (1<<3))
	LCD_DATA_PORT |= (1<<LCD_DB7);
	else
	LCD_DATA_PORT &= ~(1<<LCD_DB7);
	
	LCD_COM_PORT &= ~(1<<LCD_EN);
	_delay_ms(1);
}

//---------------------------------------------------
static void LCDSendByte(char data){
	LCDSendHalf(data>>0x04);
	LCDSendHalf(data);
	//for the lowest possible power consumption:
	LCD_DATA_PORT |= (1<<LCD_EN | 1<<LCD_RS | 1<<LCD_DB4 | 1<<LCD_DB5 | 1<<LCD_DB6 | 1<<LCD_DB7);
	//DelayMs(1);
}

//---------------------------------------------------
static void LCDCommand(char command){
	LCD_COM_PORT &= ~(1<<LCD_RS);
	LCDSendByte(command);
}

//---------------------------------------------------
static void LCDData(char data){
	LCD_COM_PORT |= (1<<LCD_RS);
	LCDSendByte(data);
}

//---------------------------------------------------
static void LCDClear(void){
	LCDCommand(LCDC_CLS);
	_delay_ms(2);
}

//---------------------------------------------------
static void LCDHome(void){
	LCDCommand(LCDC_HOME);
}

//---------------------------------------------------
static void LCDLocate(char lcd_y, char lcd_x){
	LCDCommand((lcd_y*0x40+lcd_x)|0x80);
}

//---------------------------------------------------
static void LCDSetSpecChar(char spec, char* pData)
{
	char count;
	if (spec < (MAX_SPEC_CHARS - 1))	//i.e 8-1 = 7
	{
		spec *= 8;
		LCDCommand(LCDC_CGA | spec);
		count = 0;
		while (count < 8)
		{
			LCDData(*pData);
			pData++;
			count++;
		}
	}
}
