/*
 * serial.c
 *
 * Created: 2017-07-17 00:41:14
 *  Author: doom
 */ 

//---------------------------------------------------
#define F_CPU 8000000
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <inttypes.h>
#include "serial.h"
#include "settings.h"

//-------------------------------------------
#define SERIAL_COM_ENTER			0x0D		//\r
#define SERIAL_MIN_FRAME_LEN		2

#define SERIAL_TIMDIV_STOP			0
#define SERIAL_TIMDIV_1				1
#define SERIAL_TIMDIV_8				2
#define SERIAL_TIMDIV_64			3
#define SERIAL_TIMDIV_256			4
#define SERIAL_TIMDIV_1024			5

//-------------------------------------------
struct {
	uint8_t timval;
	uint8_t timdiv;
}serialConfig;

struct {
	char wbuff[SERIAL_WBUFFER_SIZE];
	uint16_t wcounter;
	volatile uint8_t wstatus;
	volatile uint8_t *wpt;
	volatile uint16_t scounter;
	
	char rbuff[SERIAL_RBUFFER_SIZE];
	uint16_t rcounter;
	volatile char rstatus;
}serial;


static void UartInit(void);
static void TimerInit(void);
static void TimerStart(void);
static void TimerStop(void);

//---------------------------------------------------
void SerialInit(void){
	SERIAL_PORT |= (1<<SERIAL_TXD | 1<<SERIAL_RXD);
	SERIAL_DDR |= (1<<SERIAL_TXD);
	SerialBaudSet(SERIAL_BAUD_19200);
	UartInit();
	TimerInit();
}

//---------------------------------------------------
void SerialBaudSet(uint8_t baud){
	switch (baud){
		case SERIAL_BAUD_2400: {			//2400
			UBRR0H = 0;
			UBRR0L = 207;
			serialConfig.timdiv = SERIAL_TIMDIV_1024;
			serialConfig.timval = 129;			
		} break;
		case SERIAL_BAUD_4800: {			//4800
			UBRR0H = 0;
			UBRR0L = 103;
			serialConfig.timdiv = SERIAL_TIMDIV_1024;
			serialConfig.timval = 64;
		} break;
		case SERIAL_BAUD_9600: {			//9600
			UBRR0H = 0;
			UBRR0L = 51;
			serialConfig.timdiv = SERIAL_TIMDIV_1024;
			serialConfig.timval = 31;
		} break;
		case SERIAL_BAUD_14400: {			//14400
			UBRR0H = 0;
			UBRR0L = 34;
			serialConfig.timdiv = SERIAL_TIMDIV_1024;
			serialConfig.timval = 20;
		} break;
		case SERIAL_BAUD_19200: {			//19200
			UBRR0H = 0;
			UBRR0L = 25;
			serialConfig.timdiv = SERIAL_TIMDIV_256;
			serialConfig.timval = 64;
		} break;
		case SERIAL_BAUD_28800: {			//28800
			UBRR0H = 0;
			UBRR0L = 16;
			serialConfig.timdiv = SERIAL_TIMDIV_64;
			serialConfig.timval = 172;
		} break;
		case SERIAL_BAUD_57600: {			//57600
			UBRR0H = 0;
			UBRR0L = 8;
			serialConfig.timdiv = SERIAL_TIMDIV_64;
			serialConfig.timval = 85;
		} break;
		case SERIAL_BAUD_115200: {			//115200
			UBRR0H = 0;
			UBRR0L = 3;
			serialConfig.timdiv = SERIAL_TIMDIV_64;
			serialConfig.timval = 42;
		} break;
	}
}

//---------------------------------------------------
static void UartInit(void){
	UCSR0A = 0;
	UCSR0B = (1<<RXEN0 | 1<<TXEN0 | 1<<TXCIE0 | 1<<RXCIE0 | 0<<UDRIE0);		//RX
	UCSR0C = (1<<UCSZ01 | 1<<UCSZ00);				//8 data, 1 stop, parity none
}

//---------------------------------------------------
static void UartDisableRX(void){
	UCSR0B &= ~(1<<RXEN0);		//RX
}

//---------------------------------------------------
static void UartEnableRX(void){
	UCSR0B |= (1<<RXEN0);		//RX
}

//---------------------------------------------------
static void UartDisableUDREmpty(void){
	UCSR0B &= ~(1<<UDRIE0);		//RX
}

//---------------------------------------------------
static void UartEnableUDREmpty(void){
	UCSR0B |= (1<<UDRIE0);		//RX
}

//---------------------------------------------------
static void TimerInit(void){
	TCCR0A = 0;
	TCCR0B = 0;
	TIMSK0 = (1<<TOIE0);
}

//---------------------------------------------------
static void TimerStart(void){
	TCNT0 = (0xFF - serialConfig.timval);
	TCCR0B = serialConfig.timdiv;
}

//---------------------------------------------------
static void TimerStop(void){
	TCCR0B = 0;
}

//-------------------------------------------
void SerialWBufferAddChar(uint8_t data){
	if (serial.wcounter < (SERIAL_WBUFFER_SIZE - 1)) {
		serial.wbuff[serial.wcounter] = data;
		serial.wcounter++;
		serial.wbuff[serial.wcounter] = 0;
	}
}

//-------------------------------------------
uint8_t SerialWBufferAdd(uint8_t *src){
	char status = 0;
	while (status == 0) {
		if (serial.wcounter < (SERIAL_WBUFFER_SIZE - 1)) {
			if ((*src) && (src)) {
				serial.wbuff[serial.wcounter] = *src;
				serial.wcounter++;
				src++;
			} else {
				status = 1;
			}
		} else status = 2;
		
		serial.wbuff[serial.wcounter] = 0;
	}
	return status;
}

//-------------------------------------------
uint8_t SerialWBufferAddLen(uint8_t *src, uint8_t len){
	char status = 0;
	while (status == 0) {
		if (serial.wcounter < (SERIAL_WBUFFER_SIZE - 1)) {
			if (len) {
				serial.wbuff[serial.wcounter] = *src;
				serial.wcounter++;
				src++;
				len--;
			} else {
				status = 1;
			}
		} else status = 2;
		
		serial.wbuff[serial.wcounter] = 0;
	}
	return status;
}

//------------------------------------------
void SerialWBufferAddVal(int16_t val){
	uint8_t blink = 0;
	uint8_t d;
	uint8_t status = 0;
	uint16_t div = 10000;
	
	if (val < 0) {
		SerialWBufferAddChar('-');
		val *= -1;
	}
	
	//if(val > 9999) val = 9999;
	while (status == 0)
	{
		d = (val / div);
		if (d) blink = 1;
		if (div == 1) blink = 1;
		
		if (blink) {
			SerialWBufferAddChar(d + '0');
		}
		//else *pt = (' ');
		
		if (div == 1) status = 1;
		val = (val % div);
		div = (div / 10);
	}
}

//------------------------------------------
void SerialWBufferAddHexVal(int16_t val){
	uint16_t hex;
	
	SerialWBufferAddChar('x');
	
	if (val > 0xFF){
		hex = val;
		hex >>= 12;
		SerialWBufferAddChar(SerialVal2Hex(hex & 0xF));
		hex = val;
		hex >>= 8;
		SerialWBufferAddChar(SerialVal2Hex(hex & 0xF));
	}
	
	hex = val;
	hex >>= 4;
	SerialWBufferAddChar(SerialVal2Hex(hex & 0xF));
	hex = val;
	SerialWBufferAddChar(SerialVal2Hex(hex & 0xF));
}


//------------------------------------------
uint8_t SerialVal2Hex(uint8_t val){
	if (val < 10) {
		return (val + '0');
	} else {
		val -= 10;
		return (val + 'A');
	}
}

//------------------------------------------
void SerialDebug(uint8_t* src, int16_t val){
	SerialWBufferAdd("\r");
	SerialWBufferAdd(src);
	SerialWBufferAddVal(val);
}

//-------------------------------------------
void SerialWBufferSend(void){
	while (serial.wstatus > SERIAL_IDLE){};
	
	if (serial.wcounter > 0){
		UartDisableRX();
		serial.wstatus = SERIAL_SENDING;
		serial.scounter = 0;
		serial.wpt = &serial.wbuff[0];
		UDR0 = *serial.wpt;		
		TimerStart();
		UartEnableUDREmpty();
	}
}

//----------------------------------------------------
ISR(USART_UDRE_vect){
	serial.scounter++;
	if (serial.scounter < serial.wcounter){
		serial.wpt++;
		UDR0 = *serial.wpt;
		TimerStart();
	} else {
		UartDisableUDREmpty();		
	}	
}

//----------------------------------------------------
ISR(USART_TX_vect){
	UartDisableUDREmpty();		
	//UartEnableRX();
	TimerStart();
	serial.wcounter = 0;
	serial.wbuff[0] = 0;
}

//---------------------------------------------------
ISR(TIMER0_OVF_vect){
	TimerStop();

	serial.wstatus = SERIAL_IDLE;
	
	UartEnableRX();
	
	if (serial.rcounter >= SERIAL_MIN_FRAME_LEN){
		serial.rstatus = SERIAL_READY;
	} else {
		serial.rstatus = SERIAL_IDLE;
		serial.rcounter = 0;
		serial.rbuff[0] = 0;
	}
}

//-------------------------------------------
void SerialWBufferNew(void){
	while (serial.wstatus == SERIAL_SENDING){};
	serial.wcounter = 0;
	serial.wbuff[0] = 0;
}

//-------------------------------------------
void SerialWBufferQuick(uint8_t *src){
	SerialWBufferNew();
	SerialWBufferAdd(src);
	SerialWBufferSend();
}


//-------------------------------------
void SerialSendChar(uint8_t data){
	while(!(UCSR0A & 1<<UDRE0));
	UDR0 = data;
}

//-------------------------------------------
void SerialSendMsg(uint8_t *data){
	while (*data > 0){
		while(!(UCSR0A & 1<<UDRE0)){}
		UDR0 = *data;
		data++;
	}
}

//-----------------------------------
void SerialSendVal(uint16_t value){
	uint16_t temp = 10000;
	char digit;
	char blank = 1;

	while (temp)
	{
		digit = value / temp;
		value = value % temp;
		
		if (temp == 1) blank = 0;
		if (digit) blank = 0;
		
		if(!(blank)) SerialSendChar(digit + '0');
		else SerialSendChar(' ');

		temp = temp / 10;
	}
}

//----------------------------------------------------
ISR(USART_RX_vect){
	uint8_t temp;
	temp = UDR0;
	
	TimerStart();
	serial.wstatus = SERIAL_RECIVEING;	
							
	if (serial.rstatus < SERIAL_READY){
		
		//if (CfgConfigCheck() == 0) {
		if (temp != SERIAL_ENTER_CHAR) TimerStop();
		//}
		
		if (serial.rcounter < (SERIAL_RBUFFER_SIZE - 1)){
			serial.rbuff[serial.rcounter] = temp;
			serial.rcounter++;
			serial.rbuff[serial.rcounter] = 0;
		}
	}
}

//----------------------------------------------------
uint8_t SerialRBufferCheck(void){
	return serial.rstatus;
}

//----------------------------------------------------
uint8_t SerialRBufferCounter(void){
	return serial.rcounter;
}

//----------------------------------------------------
void SerialRBufferClear(void){
	serial.rstatus = SERIAL_IDLE;	
	serial.rcounter = 0;
	serial.rbuff[0] = 0;
}

//----------------------------------------------------
uint8_t SerialRBufferChar(uint8_t index){
	return serial.rbuff[index];
}

//----------------------------------------------------
uint8_t *SerialRBufferSrc(uint8_t index){
	return &serial.rbuff[index];
}



